import datetime
from dataclasses import dataclass
from typing import Optional

import click

from tecton import tecton_context
from tecton._internals import metadata_service
from tecton.cli import cli_utils
from tecton.cli import printer
from tecton.cli.command import TectonGroup
from tecton.cli.engine import EngineCommand
from tecton_core.id_helper import IdHelper
from tecton_core.specs.utils import get_timestamp_field_or_none
from tecton_proto.data import state_update__client_pb2 as state_update_pb2
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2


def format_date(datetime: datetime):
    return datetime.strftime("%Y-%m-%d %H:%M:%S %Z")


@dataclass
class PlanListItem:
    plan_id: str
    applied_by: str
    applied_at: datetime.datetime
    workspace: str
    sdk_version: str

    @classmethod
    def from_proto(cls, state_update_entry: state_update_pb2.StateUpdateEntry):
        applied_by = cli_utils.display_principal(state_update_entry.applied_by_principal, state_update_entry.applied_by)
        applied_at = get_timestamp_field_or_none(state_update_entry, "applied_at")
        return cls(
            # commit_id is called plan_id in public facing UX. Re-aliasing here.
            plan_id=state_update_entry.commit_id,
            applied_by=applied_by,
            applied_at=applied_at,
            workspace=state_update_entry.workspace,
            sdk_version=state_update_entry.sdk_version,
        )


@dataclass
class PlanSummary:
    applied_at: Optional[datetime.datetime]
    applied_by: Optional[str]
    applied: bool
    created_at: datetime.datetime
    created_by: str
    workspace: str
    sdk_version: str

    @classmethod
    def from_proto(cls, state_update_summary: state_update_pb2.StateUpdatePlanSummary):
        applied_at = get_timestamp_field_or_none(state_update_summary, "applied_at")
        applied_by = cli_utils.display_principal(
            state_update_summary.applied_by_principal, state_update_summary.applied_by
        )
        applied = bool(applied_at)
        created_at = get_timestamp_field_or_none(state_update_summary, "created_at")
        return cls(
            applied=applied,
            applied_at=applied_at,
            applied_by=applied_by,
            created_at=created_at,
            created_by=state_update_summary.created_by,
            workspace=state_update_summary.workspace,
            sdk_version=state_update_summary.sdk_version,
        )


def get_plans_list_items(workspace: str, limit: int):
    request = metadata_service_pb2.GetStateUpdateLogRequest(workspace=workspace, limit=limit)
    response = metadata_service.instance().GetStateUpdateLog(request)
    return [PlanListItem.from_proto(entry) for entry in response.entries]


def get_plan(plan_id):
    plan_id = IdHelper.from_string(plan_id)
    request = metadata_service_pb2.GetStateUpdatePlanSummaryRequest(plan_id=plan_id)
    response = metadata_service.instance().GetStateUpdatePlanSummary(request)
    return PlanSummary.from_proto(response.plan)


@click.group("plan", cls=TectonGroup)
def plan():
    r"""⚠️  Command has moved: Use `tecton plan create` to create a plan.\n
    Manage Tecton Plans in a Workspace
    """


create = EngineCommand(
    name="create",
    apply=False,
    allows_suppress_recreates=True,
    help="Create a plan: Compare your local feature definitions with remote state and show the plan to bring them in sync.",
)


plan.add_command(create)


@plan.command(uses_workspace=True)
@click.option("--limit", default=10, type=int, help="Number of log entries to return.")
def list(limit):
    """List previous tecton plans in this workspace."""
    list_impl(limit)


def list_impl(limit):
    """Implementation of tecton plan list.
    Created as a helper-function so implementation can be shared between tecton plan list
    and tecton log for backwards compatibility in 1.1
    """
    # TODO: Move implenetation into list when tecton log is removed in tecton version 1.1.0
    workspace = tecton_context.get_current_workspace()
    entries = get_plans_list_items(workspace, limit)
    table_rows = [
        (entry.plan_id, entry.applied_by, format_date(entry.applied_at), entry.sdk_version) for entry in entries
    ]
    cli_utils.display_table(["Plan Id", "Created by", "Creation Date", "SDK Version"], table_rows)


@plan.command()
@click.argument("plan-id", required=True)
def show(plan_id):
    """Show details of a Tecton Plan"""
    plan = get_plan(plan_id=plan_id)
    printer.safe_print(f"Showing current status for Plan {plan_id}")
    printer.safe_print()
    printer.safe_print(f"Created at: {format_date(plan.created_at)}")
    printer.safe_print(f"Created by: {plan.created_by}")
    printer.safe_print(f"Applied: {plan.applied}")
    if plan.applied:
        printer.safe_print(f"Applied at: {format_date(plan.applied_at)}")
        printer.safe_print(f"Applied by: {plan.applied_by}")

    # TODO: Add integration test information here



from enum import Enum


class QueryTreeStep(Enum):
    """Query trees are composed of steps.

    Each step may have its own compute, and may stage its results in preparation for the next step.

    Query trees do not have to have all steps. For example, a materialization query tree for a batch feature view will
    not have any nodes in the ODFV step.
    """

    # Runs data source scans.
    DATA_SOURCE = 1
    # Runs feature view transformations to produce un-aggregated feature data.
    PIPELINE = 2
    # Runs model inference (only used by text embeddings right now).
    MODEL_INFERENCE = 3
    # Runs partial aggregations, full aggregations, and the as-of join.
    AGGREGATION = 4
    # Runs on-demand transformations.
    ODFV = 5



from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import attrs
import pypika
from pypika import AliasedQuery
from pypika import JoinType
from pypika.functions import Cast
from pypika.functions import Coalesce
from pypika.functions import Sum
from pypika.queries import QueryBuilder
from pypika.queries import Selectable
from pypika.terms import Criterion
from pypika.terms import Field
from pypika.terms import Function
from pypika.terms import LiteralValue
from pypika.terms import Term

from tecton_core.aggregation_utils import get_aggregation_function_result_type
from tecton_core.data_types import ArrayType
from tecton_core.data_types import DataType
from tecton_core.data_types import Float32Type
from tecton_core.data_types import Float64Type
from tecton_core.data_types import Int32Type
from tecton_core.data_types import Int64Type
from tecton_core.data_types import StringType
from tecton_core.query import nodes
from tecton_core.query.aggregation_plans import AGGREGATION_PLANS
from tecton_core.query.aggregation_plans import FilterAggregationInput
from tecton_core.query.node_interface import QueryNode
from tecton_core.query.sql_compat import DuckDBTupleTerm
from tecton_core.query_consts import anchor_time
from tecton_core.query_consts import tecton_secondary_key_aggregation_indicator_col
from tecton_core.query_consts import temp_indictor_column_name
from tecton_core.specs import create_time_window_spec_from_data_proto
from tecton_core.specs.time_window_spec import LifetimeWindowSpec
from tecton_core.specs.time_window_spec import RelativeTimeWindowSpec
from tecton_core.specs.time_window_spec import TimeWindowSeriesSpec
from tecton_core.specs.time_window_spec import TimeWindowSpec
from tecton_core.time_utils import convert_timedelta_for_version


class DuckDBArray(Term):
    def __init__(self, element_type: Union[str, Term]) -> None:
        super().__init__()
        self.element_type = element_type

    def get_sql(self, **kwargs):
        element_type_sql = (
            self.element_type.get_sql(**kwargs) if isinstance(self.element_type, Term) else self.element_type
        )
        return f"{element_type_sql}[]"


DATA_TYPE_TO_DUCKDB_TYPE: Dict[DataType, str] = {
    Int32Type(): "INT32",
    Int64Type(): "INT64",
    Float32Type(): "FLOAT",
    Float64Type(): "DOUBLE",
    StringType(): "VARCHAR",
}


def _data_type_to_duckdb_type(data_type: DataType) -> Union[str, Term]:
    if not isinstance(data_type, ArrayType):
        return DATA_TYPE_TO_DUCKDB_TYPE.get(data_type, str(data_type))

    return DuckDBArray(_data_type_to_duckdb_type(data_type.element_type))


def _join_condition(left: Selectable, right: Selectable, join_keys: List[str]) -> Criterion:
    left_join_struct = DuckDBTupleTerm(*[left.field(col) for col in join_keys])
    right_join_struct = DuckDBTupleTerm(*[right.field(col) for col in join_keys])
    return left_join_struct == right_join_struct


def _column_with_default(table: Selectable, column: str, default: Optional[Term] = None) -> Term:
    if not default:
        return table[column]

    return Coalesce(table[column], default).as_(column)


class TableFunction(Term):
    def __init__(self, function: Function, table_name: str, columns: List[str]) -> None:
        self.function = function
        self.table_name = table_name
        self.columns = columns

    def get_sql(self, **kwargs):
        return f"{self.function.get_sql(**kwargs)} {self.table_name}({', '.join(self.columns)})"


@attrs.frozen
class PartialAggDuckDBNode(nodes.PartialAggNode):
    @classmethod
    def from_query_node(cls, query_node: nodes.PartialAggNode) -> QueryNode:
        return cls(
            dialect=query_node.dialect,
            compute_mode=query_node.compute_mode,
            input_node=query_node.input_node,
            fdw=query_node.fdw,
            window_start_column_name=query_node.window_start_column_name,
            aggregation_tile_interval=query_node.aggregation_tile_interval,
            window_end_column_name=query_node.window_end_column_name,
            aggregation_anchor_time=query_node.aggregation_anchor_time,
        )

    def _get_partial_agg_columns_and_names(self) -> List[Tuple[Term, str]]:
        """
        Primarily overwritten to do additional type casts to make DuckDB's post-aggregation types consistent with Spark

        The two main cases:
        - Integer SUMs: DuckDB will automatically convert all integer SUMs to cast to DuckDB INT128's, regardless
        of its original type. Note that when copying this out into parquet, DuckDB will convert these to doubles.
        - Averages: DuckDB will always widen the precision to doubles

        Spark, in contrast, maintains the same type for both of these cases
        """
        normal_agg_cols_with_names = super()._get_partial_agg_columns_and_names()
        schema = self.fdw.materialization_schema.to_dict()

        final_agg_cols = []
        for col, alias in normal_agg_cols_with_names:
            data_type = schema[alias]
            sql_type = _data_type_to_duckdb_type(data_type)

            final_agg_cols.append((Cast(col, sql_type), alias))

        return final_agg_cols


@attrs.frozen
class AsofJoinFullAggNodeDuckDBNode(nodes.AsofJoinFullAggNode):
    @classmethod
    def from_query_node(cls, query_node: nodes.AsofJoinFullAggNode) -> QueryNode:
        return cls(
            dialect=query_node.dialect,
            compute_mode=query_node.compute_mode,
            spine=query_node.spine,
            partial_agg_node=query_node.partial_agg_node,
            fdw=query_node.fdw,
            enable_spine_time_pushdown_rewrite=query_node.enable_spine_time_pushdown_rewrite,
            enable_spine_entity_pushdown_rewrite=query_node.enable_spine_entity_pushdown_rewrite,
        )

    def _get_aggregations(self, window_order_col: str, partition_cols: List[str]) -> List[Term]:
        aggregations = super()._get_aggregations(window_order_col, partition_cols)
        features = self.fdw.fv_spec.aggregate_features
        secondary_key_indicators = (
            [
                temp_indictor_column_name(secondary_key_output.time_window)
                for secondary_key_output in self.fdw.materialized_fv_spec.secondary_key_rollup_outputs
            ]
            if self.fdw.aggregation_secondary_key
            else []
        )

        view_schema = self.fdw.view_schema.to_dict()

        # (column name, column type)
        output_columns: List[Tuple["str", DataType]] = []

        for feature in features:
            input_type = view_schema[feature.input_feature_name]
            result_type = get_aggregation_function_result_type(feature.function, input_type)
            window_spec = create_time_window_spec_from_data_proto(feature.time_window)
            if isinstance(window_spec, TimeWindowSeriesSpec):
                for relative_time_window in window_spec.time_windows:
                    output_columns.append(
                        (feature.output_feature_name + "_" + relative_time_window.to_string(), result_type)
                    )
            else:
                output_columns.append((feature.output_feature_name, result_type))

        for column_name in secondary_key_indicators:
            output_columns.append((column_name, Int64Type()))

        assert len(aggregations) == len(
            output_columns
        ), "List of aggregations and list of output columns must have the same length"
        return [
            Cast(aggregation, _data_type_to_duckdb_type(tecton_type)).as_(column_name)
            for aggregation, (column_name, tecton_type) in zip(aggregations, output_columns)
        ]


@attrs.frozen
class AsofJoinFullAggDuckDBNodeV2(nodes.AsofJoinFullAggNode):
    @classmethod
    def from_query_node(cls, query_node: nodes.AsofJoinFullAggNode) -> QueryNode:
        return cls(
            dialect=query_node.dialect,
            compute_mode=query_node.compute_mode,
            spine=query_node.spine,
            partial_agg_node=query_node.partial_agg_node,
            fdw=query_node.fdw,
            enable_spine_time_pushdown_rewrite=query_node.enable_spine_time_pushdown_rewrite,
            enable_spine_entity_pushdown_rewrite=query_node.enable_spine_entity_pushdown_rewrite,
        )

    def _get_aggregation_defaults(self) -> Dict[str, Term]:
        defaults = {}
        for feature in self.fdw.fv_spec.aggregate_features:
            aggregation_plan = AGGREGATION_PLANS.get(feature.function)
            if callable(aggregation_plan):
                aggregation_plan = aggregation_plan(None, feature.function_params, False)

            defaults[feature.output_feature_name] = aggregation_plan.full_aggregation_default_value or LiteralValue(
                "null"
            )

        return defaults

    def _get_aggregation_filter_for_window_spec(
        self, window_spec: TimeWindowSpec, feature_timestamp: Term, spine_timestamp: Term
    ) -> Criterion:
        if isinstance(window_spec, RelativeTimeWindowSpec):
            range_start = convert_timedelta_for_version(
                window_spec.window_start, self.fdw.get_feature_store_format_version
            )
            range_end = convert_timedelta_for_version(window_spec.window_end, self.fdw.get_feature_store_format_version)
            filter_ = feature_timestamp.between(spine_timestamp + range_start + 1, spine_timestamp + range_end)
        elif isinstance(window_spec, LifetimeWindowSpec):
            filter_ = LiteralValue("true")
        else:
            msg = f"Aggregation window of type {window_spec.__class__} is not supported by Rift"
            raise RuntimeError(msg)

        return filter_

    def _get_aggregations(self, feature_timestamp: Term, spine_timestamp: Term) -> List[Term]:
        features = self.fdw.fv_spec.aggregate_features
        aggregation_continuous = self.fdw.trailing_time_window_aggregation().is_continuous
        aggregations = []
        view_schema = self.fdw.view_schema.to_dict()
        # (column name, column type)
        output_columns: List[Tuple["str", DataType]] = []

        for feature in features:
            aggregation_plan = AGGREGATION_PLANS.get(feature.function)
            if callable(aggregation_plan):
                aggregation_plan = aggregation_plan(feature_timestamp, feature.function_params, aggregation_continuous)

            names = aggregation_plan.materialized_column_names(feature.input_feature_name)
            window_spec = create_time_window_spec_from_data_proto(feature.time_window)
            if isinstance(window_spec, TimeWindowSeriesSpec):
                msg = "TimeWindowSeriesSpec is currently not supported for the optimized Rift query tree. To use this feature, please set conf.set('DUCKDB_ENABLE_OPTIMIZED_FULL_AGG', False)"
                raise NotImplementedError(msg)
            filter_ = self._get_aggregation_filter_for_window_spec(window_spec, feature_timestamp, spine_timestamp)

            aggregations.append(aggregation_plan.full_aggregation_with_filter_query_term(names, filter_))
            input_type = view_schema[feature.input_feature_name]
            result_type = get_aggregation_function_result_type(feature.function, input_type)
            output_columns.append((feature.output_feature_name, result_type))

        if self.fdw.aggregation_secondary_key:
            for secondary_key_output in self.fdw.materialized_fv_spec.secondary_key_rollup_outputs:
                indicator_column = temp_indictor_column_name(secondary_key_output.time_window)
                filter_ = self._get_aggregation_filter_for_window_spec(
                    secondary_key_output.time_window, feature_timestamp, spine_timestamp
                )
                aggregations.append(
                    FilterAggregationInput(
                        aggregation=Sum(Field(tecton_secondary_key_aggregation_indicator_col())), filter_=filter_
                    )
                )
                output_columns.append((indicator_column, Int64Type()))

        assert len(aggregations) == len(
            output_columns
        ), "List of aggregations and list of output columns must have the same length"
        return [
            Cast(aggregation, _data_type_to_duckdb_type(tecton_type)).as_(column_name)
            for aggregation, (column_name, tecton_type) in zip(aggregations, output_columns)
        ]

    def _augment_spine_with_join_anchor_column(self, spine: AliasedQuery, column_name: str) -> QueryBuilder:
        """
        Explode spine rows by adding all possible anchor times (expected to be equal anchors in the feature input),
        that would be needed to calculate features for this spine row.

        This new column aimed to be used in JOIN condition with feature input.

        It's computed as:
            SELECT _anchor_time + delta FROM spine CROSS JOIN deltas
        where `deltas` is
            range(earliest_window_start, 0, tile_interval)

        Example:
            Given feature view with the earliest aggregation window = -7d

            input spine:
            |  join_key  | timestamp        | anchor_time         |
            |      AAAA  | 2024-02-05 11:00 | 1707091200000000000 |

            output spine:
            |  join_key  | timestamp        | anchor_time         |  _join_anchor_time  |
            |      AAAA  | 2024-02-05 11:00 | 1707091200000000000 | 1707091200000000000 |
            |      AAAA  | 2024-02-05 11:00 | 1707091200000000000 | 1707004800000000000 |
            |      AAAA  | 2024-02-05 11:00 | 1707091200000000000 | 1706918400000000000 |
            |      AAAA  | 2024-02-05 11:00 | 1707091200000000000 | 1706745600000000000 |
            |      AAAA  | 2024-02-05 11:00 | 1707091200000000000 | 1706659200000000000 |
            |      AAAA  | 2024-02-05 11:00 | 1707091200000000000 | 1706572800000000000 |
            |      AAAA  | 2024-02-05 11:00 | 1707091200000000000 | 1706486400000000000 |
        """
        assert not self.fdw.has_lifetime_aggregate and not self.fdw.is_continuous

        earliest_window_start = convert_timedelta_for_version(
            self.fdw.earliest_window_start, self.fdw.get_feature_store_format_version
        )
        tile_interval = convert_timedelta_for_version(
            self.fdw.get_tile_interval, self.fdw.get_feature_store_format_version
        )
        deltas = TableFunction(
            function=Function("GENERATE_SERIES", earliest_window_start, 0, tile_interval),
            table_name="deltas",
            columns=["delta"],
        )
        # we need here only join keys and anchor
        core_spine_columns = [*self.fdw.join_keys, anchor_time()]
        return (
            self.func.query()
            .from_(spine)
            .join(AliasedQuery("deltas", deltas))
            .cross()
            .select(
                *(
                    [spine.field(col) for col in core_spine_columns]
                    + [(spine.field(anchor_time()) + Field("delta")).as_(column_name)]
                )
            )
        )

    def _find_all_related_spine_rows(
        self, features: AliasedQuery, spine: AliasedQuery, spine_anchor_column: str
    ) -> QueryBuilder:
        """
        Performs "reverse" join: for each row in the feature input find all spine rows that would depend on this feature
        row in the full aggregation calculation.

        The resulting table contains all columns from the feature input and the `_anchor_time` for the spine input,
        renamed into {spine_anchor_column}. The output is expected to be bigger than the input, because a single feature
        row might be involved in calculating multiple spine rows.

        For BWAFV (except ones that contain lifetime aggregations) and non-continuous SWAFV we will use HASH JOIN.
        This is achieved by augmenting spine input by adding _join_anchor_time
        (see `_augment_spine_with_join_anchor_column` method).

        For continuous SWAFV and BWAFV with lifetime aggregations we will fallback to inequality-based JOIN,
        which is expected to be less efficient.
        """
        select_cols = [features.field(col) for col in self.partial_agg_node.columns]
        if self.fdw.aggregation_secondary_key:
            select_cols += [LiteralValue("1").as_(tecton_secondary_key_aggregation_indicator_col())]

        if self.fdw.has_lifetime_aggregate or self.fdw.is_continuous:
            join_condition = _join_condition(features, spine, self.fdw.join_keys)
            join_condition &= features.field(anchor_time()) <= spine.field(anchor_time())
            if self.fdw.is_continuous:
                earliest_window_start = convert_timedelta_for_version(
                    self.fdw.earliest_window_start, self.fdw.get_feature_store_format_version
                )
                join_condition &= features.field(anchor_time()) > (spine.field(anchor_time()) + earliest_window_start)
            return (
                self.func.query()
                .from_(features)
                .join(spine)
                .on(join_condition)
                .select(*[*select_cols, spine.field(anchor_time()).as_(spine_anchor_column)])
            )

        join_anchor_column = f"_join_{anchor_time()}"
        augmented_spine = self._augment_spine_with_join_anchor_column(spine, join_anchor_column)

        join_condition = _join_condition(features, augmented_spine, self.fdw.join_keys)
        join_condition &= features.field(anchor_time()) == augmented_spine.field(join_anchor_column)

        return (
            self.func.query()
            .from_(features)
            .join(augmented_spine)
            .on(join_condition)
            .select(*[*select_cols, augmented_spine.field(anchor_time()).as_(spine_anchor_column)])
        )

    def _to_query(self) -> QueryBuilder:
        """
        DuckDB-optimized query for Full Aggregation calculation.
        It's based on (1) JOINs instead of UNION and (2) FILTERing aggregation input instead of WINDOWing
        (see https://duckdb.org/docs/sql/query_syntax/filter.html for more details).

        The very high-level query plan looks like following:
        1. Explode spine to add all anchor_times (matching anchors in the feature input) required to calculate
           each spine row. (doesn't apply to continuous SWAFV and BWAFV with lifetime aggregations)

        2. Join the feature input with exploded spine: essentially assigning feature row to rows in the spine
          (rows from the feature input can be duplicated at this point).

        3. Group the feature input by join keys + assigned spine anchor time column.

        4. Calculate full aggregations w/ final timestamp filtering on column-level:
            SUM(sum_impression) FILTER (features._anchor_time <= spine._anchor_time AND
                                        features._anchor_time > spine._anchor_time - epoch_ns(INTERVAL 1 DAY))
                AS sum_impression_1d,
            SUM(sum_impression) FILTER (features._anchor_time <= spine._anchor_time AND
                                        features._anchor_time > spine._anchor_time - epoch_ns(INTERVAL 7 DAY))
                AS sum_impression_7d,
            SUM(sum_impression) FILTER (features._anchor_time <= spine._anchor_time - epoch_ns(INTERVAL 2 DAY) AND
                                        features._anchor_time > spine._anchor_time - epoch_ns(INTERVAL 7 DAY))
                AS sum_impression_offset_2d_7d
        """
        join_keys = self.fdw.join_keys
        if self.fdw.aggregation_secondary_key:
            join_keys += [self.fdw.aggregation_secondary_key]

        spine_anchor_time = f"_spine_{anchor_time()}"
        partition_cols = [*join_keys, spine_anchor_time]

        features_alias = self.partial_agg_node.name
        spine_alias = self.spine.name
        features_q = AliasedQuery(features_alias)
        spine_q = AliasedQuery(spine_alias)

        features_with_spine_anchor = self._find_all_related_spine_rows(features_q, spine_q, spine_anchor_time)
        aggregations = self._get_aggregations(
            spine_timestamp=features_with_spine_anchor[spine_anchor_time],
            feature_timestamp=features_with_spine_anchor[anchor_time()],
        )

        calculated_features = (
            self.func.query()
            .from_(features_with_spine_anchor)
            .groupby(*partition_cols)
            .select(*([features_with_spine_anchor.field(col) for col in partition_cols] + aggregations))
        )

        join_condition = _join_condition(spine_q, calculated_features, join_keys)
        join_condition &= spine_q[anchor_time()] == calculated_features[spine_anchor_time]

        aggregation_defaults = self._get_aggregation_defaults()

        return (
            self.func.query()
            .with_(self.spine.node._to_query(), spine_alias, materialized=True)
            .with_(self.partial_agg_node._to_query(), features_alias)
            .from_(spine_alias)
            .left_join(calculated_features)
            .on(join_condition)
            .select(
                *(
                    [spine_q[col] for col in self.spine.node.columns]
                    + [
                        _column_with_default(calculated_features, col, aggregation_defaults.get(col))
                        for col in self.columns
                        if col not in self.spine.node.columns
                    ]
                )
            )
        )


class AsofJoin(pypika.queries.JoinOn):
    def get_sql(self, **kwargs):
        return "ASOF " + super().get_sql(**kwargs)


@attrs.frozen
class AsofJoinDuckDBNode(nodes.AsofJoinNode):
    @classmethod
    def from_query_node(cls, query_node: nodes.AsofJoinNode) -> QueryNode:
        kwargs = attrs.asdict(query_node, recurse=False)
        del kwargs["node_id"]
        return cls(**kwargs)

    def _to_query(self) -> pypika.queries.QueryBuilder:
        left_df = self.left_container.node._to_query()
        right_df = self.right_container.node._to_query()
        right_name = self.right_container.node.name
        left_cols = list(self.left_container.node.columns)
        right_none_join_cols = [col for col in self.right_container.node.columns if col not in self.join_cols]
        columns = [left_df.field(col) for col in left_cols] + [
            AliasedQuery(right_name).field(col).as_(f"{self.right_container.prefix}_{col}")
            for col in right_none_join_cols
        ]
        # Using struct here to handle null values in the join columns
        left_join_struct = DuckDBTupleTerm(*[left_df.field(col) for col in self.join_cols])
        right_join_struct = DuckDBTupleTerm(*[AliasedQuery(right_name).field(col) for col in self.join_cols])
        # We need to use both the effective timestamp and the timestamp for the join condition, as the effective timestamp can be same for multiple rows
        left_join_condition_struct = DuckDBTupleTerm(
            left_df.field(self.left_container.timestamp_field), left_df.field(self.left_container.timestamp_field)
        )
        right_join_condition_struct = DuckDBTupleTerm(
            AliasedQuery(right_name).field(self.right_container.effective_timestamp_field),
            AliasedQuery(right_name).field(self.right_container.timestamp_field),
        )

        res = self.func.query().with_(right_df, right_name).select(*columns).from_(left_df)
        # do_join doesn't return new query, but updates in-place
        res.do_join(
            AsofJoin(
                item=AliasedQuery(right_name),
                how=JoinType.left,
                criteria=Criterion.all(
                    [left_join_condition_struct >= right_join_condition_struct, left_join_struct == right_join_struct]
                ),
            )
        )

        return res



import dataclasses
import datetime
from typing import Tuple

import attrs
import pypika

from tecton_core import offline_store
from tecton_core import query_consts
from tecton_core.compute_mode import ComputeMode
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query.dialect import Dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.node_interface import QueryNode
from tecton_proto.data.feature_store__client_pb2 import FeatureStoreFormatVersion


@dataclasses.dataclass
class TimeSpec:
    timestamp_key: str
    # Can be either the anchor time or timestamp column depending on FeatureView type
    time_column: str
    partition_size: datetime.timedelta
    partition_is_anchor: bool

    @staticmethod
    def for_feature_definition(fd: FeatureDefinitionWrapper) -> "TimeSpec":
        return TimeSpec(
            timestamp_key=fd.timestamp_key,
            partition_size=offline_store.partition_size_for_delta(fd).as_timedelta(),
            time_column=query_consts.anchor_time() if fd.is_temporal_aggregate else fd.timestamp_key,
            partition_is_anchor=fd.is_temporal_aggregate,
        )


@attrs.frozen
class AddTimePartitionNode(QueryNode):
    input_node: NodeRef
    time_spec: TimeSpec

    @property
    def columns(self) -> Tuple[str, ...]:
        return (*self.input_node.columns, offline_store.TIME_PARTITION)

    @property
    def inputs(self) -> Tuple[NodeRef]:
        return (self.input_node,)

    def as_str(self) -> str:
        return f"Partition by {self.time_spec.time_column} in buckets of {self.time_spec.partition_size}"

    def _to_query(self) -> pypika.queries.QueryBuilder:
        partition_size_seconds = offline_store.window_size_seconds(self.time_spec.partition_size)
        assert partition_size_seconds != 0, "Can't partition by 0-width buckets"
        if self.time_spec.partition_is_anchor:
            # Anchor time is ns
            epoch_seconds = self.func.int_div(pypika.Field(self.time_spec.time_column), 1_000_000_000)
        else:
            epoch_seconds = self.func.to_unixtime(pypika.Field(self.time_spec.time_column))
        partition_epoch = epoch_seconds - (epoch_seconds % partition_size_seconds)
        ts_format = offline_store.timestamp_formats(self.time_spec.partition_size).python_format
        partition_col = self.func.strftime(self.func.from_unixtime(partition_epoch), ts_format).as_(
            offline_store.TIME_PARTITION
        )
        return pypika.Query().from_(self.input_node._to_query()).select("*", partition_col)

    @staticmethod
    def for_feature_definition(fd: FeatureDefinitionWrapper, input_node: NodeRef) -> NodeRef:
        assert (
            fd.get_feature_store_format_version
            >= FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_TIME_NANOSECONDS
        ), f"FeatureStoreFormateVersion {fd.get_feature_store_format_version} is not supported"
        return AddTimePartitionNode(
            dialect=Dialect.DUCKDB,
            compute_mode=ComputeMode.RIFT,
            input_node=input_node,
            time_spec=TimeSpec.for_feature_definition(fd),
        ).as_ref()



import base64
import glob
import io
import json
import logging
import os
from abc import ABC
from abc import abstractmethod
from typing import Optional
from typing import Union
from urllib.parse import urlparse

import boto3
import pandas
import pyspark
from google.protobuf.json_format import MessageToJson
from pyspark.sql.types import StructType

from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals.display import Displayable
from tecton._internals.sdk_decorators import sdk_public_method
from tecton.framework.data_frame import TectonDataFrame
from tecton.tecton_context import TectonContext
from tecton_core.id_helper import IdHelper
from tecton_proto.data.saved_feature_data_frame__client_pb2 import SavedFeatureDataFrame
from tecton_proto.data.saved_feature_data_frame__client_pb2 import SavedFeatureDataFrameType
from tecton_proto.metadataservice.metadata_service__client_pb2 import ArchiveSavedFeatureDataFrameRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import CreateSavedFeatureDataFrameRequest
from tecton_spark.spark_schema_wrapper import SparkSchemaWrapper


logger = logging.getLogger(__name__)


class Dataset(ABC):
    """
    Abstract base class for Dataset, representing a dataset with various metadata
    and the ability to convert to a dataframe.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Name of the dataset.
        """

    @property
    @abstractmethod
    def is_archived(self) -> bool:
        """
        Boolean indicating if the dataset is archived.
        """

    @abstractmethod
    def to_pandas(self) -> pandas.DataFrame:
        """
        Converts the Dataset to a Pandas DataFrame and returns it.
        """

    @sdk_public_method
    def to_spark(self) -> pyspark.sql.DataFrame:
        """
        Converts the Dataset to a Spark DataFrame and returns it.
        """


class SavedDataset(Dataset):
    pass


class LegacyDataset(Dataset):
    """
    Old Dataset class.

    Persisted data consisting of entity & request keys, timestamps, and calculated features. Datasets are
    associated with either a :class:`FeatureService` or :class:`FeatureView`.

    There are 2 types of Datasets: Saved and Logged.

    Saved Datasets are generated manually when calling :meth:`tecton.get_historical_features` by setting the ``save``
    parameter.

    Logged Datasets are generated automatically when declaring a :class:`FeatureService` with :class:`tecton.LoggingConfig`,
    and the data is continuously added to it when requesting online data from the FeatureService.

    To get an existing Dataset, call :py:meth:`tecton.get_dataset`.
    """

    _proto: SavedFeatureDataFrame = None
    _tecton_df: TectonDataFrame = None

    def __init__(self, proto: SavedFeatureDataFrame, spark_df: pyspark.sql.DataFrame, pandas_df: pandas.DataFrame):
        self._tecton_df = TectonDataFrame(spark_df, pandas_df, snowflake_df=None)
        self._proto = proto

    @classmethod
    def _from_proto(cls, proto):
        return cls(proto, None, None)

    @sdk_public_method
    def to_spark(self) -> pyspark.sql.DataFrame:
        """Converts the Dataset to a Spark DataFrame and returns it."""
        self._try_fetch_spark_df()
        return self._tecton_df.to_spark()

    @sdk_public_method
    def to_pandas(self) -> pandas.DataFrame:
        """Converts the Dataset to a Pandas DataFrame and returns it."""
        if self._tecton_df._pandas_df is not None:
            return self._tecton_df._pandas_df

        self._try_fetch_spark_df()
        return self._tecton_df.to_pandas()

    @sdk_public_method
    def fetch_as_pandas(self, n_samples: Optional[int] = None, **kwargs) -> pandas.DataFrame:
        """
        Fetches a saved dataset from S3 as a pandas DataFrame.

        :param n_samples: Number of samples to read from parquet files. If None, read all.
        :param kwargs: Additional arguments to pass to pandas.read_parquet function.
        :return: pandas DataFrame containing the saved dataset
        """
        # TODO: support Logged datasets and mimic how we use spark schema to correct avro types
        if self._type == SavedFeatureDataFrameType.LOGGED:
            raise errors.UNSUPPORTED_FETCH_AS_PANDAS_AVRO

        # TODO: allow other storage options for datasets
        # at the moment, Datasets are created with s3 as the only option
        # local path support is just for debugging
        if self._path.startswith("s3://"):
            o = urlparse(self._path, allow_fragments=False)
            s3_bucket, path_prefix = o.netloc, o.path.lstrip("/")
            # TODO: assumes user has local aws creds setup
            s3_client = boto3.client("s3")
            s3 = boto3.resource("s3")

            keys = [
                item.key
                for item in s3.Bucket(s3_bucket).objects.filter(Prefix=path_prefix)
                if item.key.endswith(".parquet")
            ]

            def read_func(key):
                obj = s3_client.get_object(Bucket=s3_bucket, Key=key)
                return pandas.read_parquet(io.BytesIO(obj["Body"].read()), **kwargs)

        elif os.path.exists(self._path):
            keys = glob.glob(os.path.join(self._path, "*.parquet"))

            def read_func(key):
                return pandas.read_parquet(key, **kwargs)

        else:
            raise errors.INVALID_DATASET_PATH(self._path)

        if not keys:
            logger.warning(f"Dataset {self.name} does not have any materialized data in {self.storage_location}.")
            schema = self._get_schema()
            return pandas.DataFrame(columns=[field.name for field in schema.fields])

        def read_parquet_files(keys, n_samples):
            total_yielded = 0
            for key in keys:
                if n_samples is not None and total_yielded >= n_samples:
                    break
                df = read_func(key)
                rows_to_yield = (n_samples - total_yielded) if n_samples is not None else None
                yield df.iloc[:rows_to_yield] if rows_to_yield else df
                total_yielded += len(df)

        self._tecton_df._pandas_df = pandas.concat(read_parquet_files(keys, n_samples), ignore_index=True)
        return self._tecton_df._pandas_df

    # Creates and returns an empty Spark dataframe & pandas dataframe with desired schema
    def _create_empty_dfs(self):
        schema = self._get_schema()
        spark = TectonContext.get_instance()._get_spark()

        spark_df = spark.createDataFrame(spark.sparkContext.emptyRDD(), schema)
        pandas_df = pandas.DataFrame(columns=[field.name for field in schema.fields])
        return spark_df, pandas_df

    def _get_schema(self) -> StructType:
        schema_json = json.loads(MessageToJson(self._proto.schema))
        fields = []
        for field in schema_json["fields"]:
            fields.append(json.loads(field["structfieldJson"]))

        return StructType.fromJson({"fields": fields})

    # Tries fetching self._tecton_df._spark_df. As long as the underlying data exists,
    # it's expected to succeed. However, in certain cases self._tecton_df._spark_df may stay None.
    # For example, if this is a logged dataset and there are not feature requests logged
    # yet, self._tecton_df._spark_df will stay None after the execution of this method.
    def _try_fetch_spark_df(self):
        if self._tecton_df._spark_df is not None:
            return
        spark = TectonContext.get_instance()._get_spark()
        try:
            if self._type == SavedFeatureDataFrameType.LOGGED:
                # Logged datasets are in Avro format
                self._tecton_df._spark_df = spark.read.format("avro").load(self._path)
                self._tecton_df._spark_df = _convert_logged_df_schema(self._tecton_df._spark_df)
            else:
                self._tecton_df._spark_df = spark.read.parquet(self._path)
        except pyspark.sql.utils.AnalysisException as e:
            # If the path doesn't exist in S3, there is no data
            # This can happen for logged features when there is no logs yet,
            # so we don't want to throw an error in this case
            if "Path does not exist" in e.desc:
                self._tecton_df._spark_df, self._tecton_df._pandas_df = self._create_empty_dfs()
            else:
                raise e

    @sdk_public_method
    def summary(self) -> Displayable:
        """
        Print out a summary of this class's attributes.
        """
        return Displayable.from_properties(items=self._summary_items())

    def _summary_items(self):
        items = [
            ("Name", self.name),
            ("Id", IdHelper.to_string(self._proto.saved_feature_dataframe_id)),
            ("Created At", self._proto.info.created_at.ToJsonString()),
            ("Workspace", self._proto.info.workspace or "prod"),
            ("Tecton Log Commit Id", self._proto.state_update_entry_commit_id),
            ("Type", "Logged" if self._type == SavedFeatureDataFrameType.LOGGED else "Saved"),
        ]
        items.append(self._get_source())
        if len(self._proto.join_key_column_names) > 0:
            items.append(("Join & Request Keys", ", ".join(self._proto.join_key_column_names)))
        if self._proto.HasField("timestamp_column_name"):
            items.append(("Timestamp Key", self._proto.timestamp_column_name))
        return items

    def _get_source(self):
        if self._proto.HasField("feature_package_name"):
            return ("Source FeatureView", self._proto.feature_package_name)
        elif self._proto.HasField("feature_service_name"):
            return ("Source FeatureService", self._proto.feature_service_name)
        else:
            # should be unreachable
            assert False, "Neither feature_package_name nor feature_service_name set in the proto"

    def _delete(self):
        """
        Delete this Dataset. Note that this deletes the underlying data as well as removing the Dataset object from
        Tecton.
        """
        request = ArchiveSavedFeatureDataFrameRequest()
        request.saved_feature_dataframe_id.CopyFrom(IdHelper.from_string(self._id))
        metadata_service.instance().ArchiveSavedFeatureDataFrame(request)
        logger.info(f"Dataset {self.name} deleted")

    @sdk_public_method
    def get_spine_dataframe(self) -> TectonDataFrame:
        logger.warning(errors.GET_SPINE_DF_DEPRECATED)
        return self.get_events_dataframe()

    @sdk_public_method
    def get_events_dataframe(self) -> TectonDataFrame:
        """
        Get a :py:class:`tecton.TectonDataFrame` containing the spine.
        """
        if not (self._proto.join_key_column_names and self._proto.timestamp_column_name):
            raise errors.DATASET_SPINE_COLUMNS_NOT_SET

        if self._tecton_df._pandas_df is not None:
            spine_pandas_df = self._tecton_df._pandas_df[
                self._proto.join_key_column_names[:] + [self._proto.timestamp_column_name]
            ].copy()
            return TectonDataFrame(None, spine_pandas_df)

        self._try_fetch_spark_df()
        spine_spark_df = self._tecton_df._spark_df.select(
            self._proto.join_key_column_names[:] + [self._proto.timestamp_column_name]
        )
        return TectonDataFrame(spine_spark_df, None)

    @property
    def name(self):
        """
        Dataset name
        """
        return self._proto.info.name

    @property
    def is_archived(self) -> bool:
        """
        Whether the dataset record is archived.
        Stored data associated with archived datasets will be cleaned up.
        """
        return self._proto.info.is_archived

    @property
    def storage_location(self):
        """
        Dataset storage location
        """
        return self._path

    @property
    def _id(self):
        return IdHelper.to_string(self._proto.saved_feature_dataframe_id)

    @property
    def _path(self):
        return self._proto.dataframe_location

    @property
    def _feature_service_id(self):
        return IdHelper.to_string(self._proto.feature_service_id)

    @property
    def _type(self):
        return self._proto.type

    @classmethod
    def _create(
        cls,
        df: TectonDataFrame,
        save_as: Optional[str] = None,
        workspace: Optional[str] = None,
        feature_definition_id: Optional[str] = None,
        feature_service_id: Optional[str] = None,
        spine: Optional[Union[pyspark.sql.dataframe.DataFrame, pandas.DataFrame]] = None,
        timestamp_key: Optional[str] = None,
    ) -> TectonDataFrame:
        assert (
            feature_definition_id or feature_service_id
        ), "Either feature_definition_id or feature_service_id must be provided"

        create_request = CreateSavedFeatureDataFrameRequest()
        if save_as:
            create_request.name = save_as
        create_request.workspace = workspace
        if feature_definition_id:
            create_request.feature_package_id.CopyFrom(IdHelper.from_string(feature_definition_id))
        if feature_service_id:
            create_request.feature_service_id.CopyFrom(IdHelper.from_string(feature_service_id))
        if spine is not None:
            # note that spine.columns works for both pandas and spark dfs
            spine_cols = [s for s in spine.columns if s != timestamp_key]
            create_request.join_key_column_names.extend(spine_cols)
            if timestamp_key is not None:  # timestamp_key can be none for odfvs
                create_request.timestamp_column_name = timestamp_key
        create_request.schema.CopyFrom(SparkSchemaWrapper.from_spark_schema(df.to_spark().schema))
        response = metadata_service.instance().CreateSavedFeatureDataFrame(create_request)
        proto = response.saved_feature_dataframe
        logger.info(f"Saved Dataset with name {proto.info.name}")
        logger.debug(f"Path is {proto.dataframe_location}")

        df.to_spark().write.save(path=proto.dataframe_location, format="parquet", mode="errorifexists")
        return cls(proto, df._spark_df, df._pandas_df)

    def __repr__(self):
        source_type, source_value = self._get_source()
        source_str = f"{source_type}='{source_value}'"
        return (
            f"{type(self).__name__}(name='{self.name}', "
            + f"{source_str}, created_at='{self._proto.info.created_at.ToJsonString()}')"
        )


def _convert_logged_df_schema(spark_df: pyspark.sql.DataFrame):
    if spark_df is None:
        return
    # Note: _partition column is not used right now, but in future
    # it can be used to optimize time-range access of this dataframe
    spark_df = spark_df.drop("_partition")
    # Note: the rest of the column names are base16 encoded due to strict
    # Avro column name validation (only [_a-zA-Z0-9] allowed). The encoding
    # happens here:
    for column in spark_df.columns:
        new_column = base64.b16decode(column[1:], casefold=True).decode()
        spark_df = spark_df.withColumnRenamed(column, new_column)
    return spark_df
